/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is NS 4.x hooks for mozilla.
 *
 * The Initial Developer of the Original Code is
 * Christian Biesinger.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include <objbase.h>
#include <winerror.h>

#include "ns4Hooks.h"
#include "nsIURI.h"
#include "nsString.h"
#include "nsAutoPtr.h"

NS_IMPL_ISUPPORTS1(ns4Hooks, nsIContentPolicy)

ns4Hooks::~ns4Hooks() {
  CoUninitialize();
}

nsresult ns4Hooks::Init() {
  HRESULT res = CoInitialize(NULL);
  if (FAILED(res))
    return NS_ERROR_FAILURE;

  if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Netscape\\Netscape Navigator\\Automation Protocols",
                   0, KEY_QUERY_VALUE, &mRegKey) != ERROR_SUCCESS)
    return NS_ERROR_FAILURE;

  return NS_OK;
}

// Private methods
IDispatch* ns4Hooks::GetDispatchForProtocol(nsAFlatCString& aProtocol) {
  if (aProtocol.IsEmpty())
    return nsnull;

  // First, find the size of this entry so we can allocate memory...
  DWORD len;
  LONG res = RegQueryValueEx(mRegKey, aProtocol.get(), NULL, NULL, NULL, &len);
  if (res != ERROR_SUCCESS)
    return nsnull;

  nsAutoPtr<char> progid = new char[len + 1];
  if (!progid)
    return nsnull;

  res = RegQueryValueEx(mRegKey, aProtocol.get(), NULL, NULL, (LPBYTE)progid.get(), &len);
  if (res != ERROR_SUCCESS)
    return nsnull;

  CLSID dlmgrCLS = CLSID_NULL;
  HRESULT hres = CLSIDFromProgID(NS_ConvertASCIItoUTF16(progid).get(), &dlmgrCLS);
  if (FAILED(hres))
    return nsnull;

  IDispatch* pDispatch = NULL;
  res = CoCreateInstance(dlmgrCLS, NULL, CLSCTX_ALL, IID_IDispatch, (void**)&pDispatch);

  return pDispatch;
}

short ns4Hooks::CallInitialize(IDispatch *aDispatch, const nsAFlatCString& aScheme,
                               const nsAFlatCString& aSpec) {
  // OK... prepare for invocation
  DISPPARAMS initParams;
  initParams.cArgs = 2;
  initParams.cNamedArgs = 0;
  initParams.rgvarg = new VARIANTARG[2];
  if (!initParams.rgvarg)
    return 0; // don't handle it
  initParams.rgdispidNamedArgs = NULL;
  // Parameters will be stored in reverse order, just to annoy people...
  // Param 1: protocol
  VariantInit(initParams.rgvarg);
  initParams.rgvarg[1].vt = VT_BSTR;
  initParams.rgvarg[1].bstrVal = SysAllocString(NS_ConvertUTF8toUTF16(aScheme).get());
  if (!initParams.rgvarg[1].bstrVal) {
    delete[] initParams.rgvarg;
    return 0;
  }

  // Param 2: URL
  VariantInit(initParams.rgvarg + 1);
  initParams.rgvarg[0].vt = VT_BSTR;
  initParams.rgvarg[0].bstrVal = SysAllocString(NS_ConvertUTF8toUTF16(aSpec).get());
  if (!initParams.rgvarg[0].bstrVal) {
    SysFreeString(initParams.rgvarg[1].bstrVal);
    delete[] initParams.rgvarg;
    return 0;
  }

  // invoke
  VARIANT result;
  VariantInit(&result);
  UINT argerr;
  // Init has DISPID of 1
  HRESULT res = aDispatch->Invoke(1, IID_NULL, LOCALE_SYSTEM_DEFAULT,
                                  DISPATCH_METHOD, &initParams, &result,
                                  NULL, &argerr);

  // Clean up
  VariantClear(initParams.rgvarg);
  VariantClear(initParams.rgvarg + 1);
  delete[] initParams.rgvarg;

  if (SUCCEEDED(res) && result.vt == VT_I2)
    return result.intVal;

  // On failure, return 0... don't handle it
  return 0;
}

nsresult ns4Hooks::CallOpen(IDispatch* aDispatch, const nsAFlatCString& aSpec) {
  // Let it actually download...
  DISPPARAMS param;
  param.cArgs = 1;
  param.cNamedArgs = 0;
  param.rgvarg = new VARIANTARG;
  VariantInit(param.rgvarg);
  param.rgvarg[0].vt = VT_BSTR;
  param.rgvarg[0].bstrVal = SysAllocString(NS_ConvertUTF8toUTF16(aSpec).get());
  if (!param.rgvarg[0].bstrVal) {
    delete param.rgvarg;
    return NS_ERROR_OUT_OF_MEMORY;
  }

  param.rgdispidNamedArgs = NULL;

  // Open has DISPID of 2
  UINT argerr;
  HRESULT res = aDispatch->Invoke(2, IID_NULL, LOCALE_SYSTEM_DEFAULT,
                                  DISPATCH_METHOD, &param, NULL, NULL,
                                  &argerr);

  VariantClear(param.rgvarg);
  delete param.rgvarg;

  return SUCCEEDED(res) ? NS_OK : NS_ERROR_FAILURE;
}

// nsIContentPolicy implementation
NS_IMETHODIMP ns4Hooks::ShouldLoad(PRInt32 aContentType, 
                                   nsIURI *aURI,
                                   nsISupports *aContext,
                                   nsIDOMWindow *aWindow,
                                   PRBool *aShouldLoad) {
  *aShouldLoad = PR_TRUE;

  if (aContentType == nsIContentPolicy::DOCUMENT) {
    nsCAutoString protocol;
    nsresult rv = aURI->GetScheme(protocol);
    if (NS_FAILED(rv))
      return NS_OK; // return and handle normally
    IDispatch* pDispatch = GetDispatchForProtocol(protocol);
    // If we couldn't get a dispatch interface, return
    if (!pDispatch)
      return NS_OK;

    nsCAutoString spec;
    rv = aURI->GetAsciiSpec(spec);
    if (NS_FAILED(rv)) {
      pDispatch->Release();
      return NS_OK; // nothing we can do here
    }
    short handle = CallInitialize(pDispatch, protocol, spec);
    // 0=not handle, 1=handle
    if (handle != 1) {
      pDispatch->Release();
      return NS_OK;
    }

    // Ask the handler to handle the url
    if (NS_SUCCEEDED(CallOpen(pDispatch, spec)))
      *aShouldLoad = PR_FALSE;

    pDispatch->Release();
  }
  return NS_OK;
}

NS_IMETHODIMP ns4Hooks::ShouldProcess(PRInt32 aContentType,
                                      nsIURI *aDocumentLoc,
                                      nsISupports *aContext,
                                      nsIDOMWindow *aWindow,
                                      PRBool *_retval) {
  *_retval = PR_TRUE;
  return NS_OK;
}

